@@ -1,6 +1,5 @@
 use clap::ArgMatches;
 use dotenv::dotenv;
-use log::LevelFilter;
 use sea_orm_codegen::{EntityTransformer, OutputFile, WithSerde};
 use std::{error::Error, fmt::Display, fs, io::Write, path::Path, process::Command, str::FromStr};
 use url::Url;
@@ -17,6 +16,7 @@ async fn main() {
         ("generate", Some(matches)) => run_generate_command(matches)
             .await
             .unwrap_or_else(handle_error),
+        ("migrate", Some(matches)) => run_migrate_command(matches).unwrap_or_else(handle_error),
         _ => unreachable!("You should never see this message"),
     }
 }
@@ -33,9 +33,9 @@ async fn run_generate_command(matches: &ArgMatches<'_>) -> Result<(), Box<dyn Er
             let expanded_format = args.is_present("EXPANDED_FORMAT");
             let with_serde = args.value_of("WITH_SERDE").unwrap();
             if args.is_present("VERBOSE") {
-                let _ = ::env_logger::builder()
-                    .filter_level(LevelFilter::Debug)
-                    .is_test(true)
+                let _ = tracing_subscriber::fmt()
+                    .with_max_level(tracing::Level::DEBUG)
+                    .with_test_writer()
                     .try_init();
             }
 
@@ -51,42 +51,19 @@ async fn run_generate_command(matches: &ArgMatches<'_>) -> Result<(), Box<dyn Er
             // Missing scheme will have been caught by the Url::parse() call
             // above
             let url_username = url.username();
-            let url_password = url.password();
             let url_host = url.host_str();
 
-            // Panic on any that are missing
-            if url_username.is_empty() {
-                panic!("No username was found in the database url");
-            }
-            if url_password.is_none() {
-                panic!("No password was found in the database url");
-            }
-            if url_host.is_none() {
-                panic!("No host was found in the database url");
-            }
+            let is_sqlite = url.scheme() == "sqlite";
 
-            // The database name should be the first element of the path string
-            //
-            // Throwing an error if there is no database name since it might be
-            // accepted by the database without it, while we're looking to dump
-            // information from a particular database
-            let database_name = url
-                .path_segments()
-                .unwrap_or_else(|| {
-                    panic!(
-                        "There is no database name as part of the url path: {}",
-                        url.as_str()
-                    )
-                })
-                .next()
-                .unwrap();
-
-            // An empty string as the database name is also an error
-            if database_name.is_empty() {
-                panic!(
-                    "There is no database name as part of the url path: {}",
-                    url.as_str()
-                );
+            // Skip checking if it's SQLite
+            if !is_sqlite {
+                // Panic on any that are missing
+                if url_username.is_empty() {
+                    panic!("No username was found in the database url");
+                }
+                if url_host.is_none() {
+                    panic!("No host was found in the database url");
+                }
             }
 
             // Closures for filtering tables
@@ -105,6 +82,36 @@ async fn run_generate_command(matches: &ArgMatches<'_>) -> Result<(), Box<dyn Er
                 }
             };
 
+            let database_name = if !is_sqlite {
+                // The database name should be the first element of the path string
+                //
+                // Throwing an error if there is no database name since it might be
+                // accepted by the database without it, while we're looking to dump
+                // information from a particular database
+                let database_name = url
+                    .path_segments()
+                    .unwrap_or_else(|| {
+                        panic!(
+                            "There is no database name as part of the url path: {}",
+                            url.as_str()
+                        )
+                    })
+                    .next()
+                    .unwrap();
+
+                // An empty string as the database name is also an error
+                if database_name.is_empty() {
+                    panic!(
+                        "There is no database name as part of the url path: {}",
+                        url.as_str()
+                    );
+                }
+
+                database_name
+            } else {
+                Default::default()
+            };
+
             let table_stmts = match url.scheme() {
                 "mysql" => {
                     use sea_schema::mysql::discovery::SchemaDiscovery;
@@ -121,6 +128,21 @@ async fn run_generate_command(matches: &ArgMatches<'_>) -> Result<(), Box<dyn Er
                         .map(|schema| schema.write())
                         .collect()
                 }
+                "sqlite" => {
+                    use sea_schema::sqlite::SchemaDiscovery;
+                    use sqlx::SqlitePool;
+
+                    let connection = SqlitePool::connect(url.as_str()).await?;
+                    let schema_discovery = SchemaDiscovery::new(connection);
+                    let schema = schema_discovery.discover().await?;
+                    schema
+                        .tables
+                        .into_iter()
+                        .filter(|schema| filter_tables(&schema.name))
+                        .filter(|schema| filter_hidden_tables(&schema.name))
+                        .map(|schema| schema.write())
+                        .collect()
+                }
                 "postgres" | "postgresql" => {
                     use sea_schema::postgres::discovery::SchemaDiscovery;
                     use sqlx::PgPool;
@@ -166,6 +188,81 @@ async fn run_generate_command(matches: &ArgMatches<'_>) -> Result<(), Box<dyn Er
     Ok(())
 }
 
+fn run_migrate_command(matches: &ArgMatches<'_>) -> Result<(), Box<dyn Error>> {
+    let migrate_subcommand = matches.subcommand();
+    // If it's `migrate init`
+    if let ("init", Some(args)) = migrate_subcommand {
+        let migration_dir = args.value_of("MIGRATION_DIR").unwrap();
+        let migration_dir = match migration_dir.ends_with('/') {
+            true => migration_dir.to_string(),
+            false => format!("{}/", migration_dir),
+        };
+        println!("Initializing migration directory...");
+        macro_rules! write_file {
+            ($filename: literal) => {
+                write_file!($filename, $filename);
+            };
+            ($filename: literal, $template: literal) => {
+                let filepath = [&migration_dir, $filename].join("");
+                println!("Creating file `{}`", filepath);
+                let path = Path::new(&filepath);
+                let prefix = path.parent().unwrap();
+                fs::create_dir_all(prefix).unwrap();
+                let mut file = fs::File::create(path)?;
+                let content = include_str!(concat!("../template/migration/", $template));
+                file.write_all(content.as_bytes())?;
+            };
+        }
+        write_file!("src/lib.rs");
+        write_file!("src/m20220101_000001_create_table.rs");
+        write_file!("src/main.rs");
+        write_file!("Cargo.toml", "_Cargo.toml");
+        write_file!("README.md");
+        println!("Done!");
+        // Early exit!
+        return Ok(());
+    }
+    let (subcommand, migration_dir, steps, verbose) = match migrate_subcommand {
+        // Catch all command with pattern `migrate xxx`
+        (subcommand, Some(args)) => {
+            let migration_dir = args.value_of("MIGRATION_DIR").unwrap();
+            let steps = args.value_of("NUM_MIGRATION");
+            let verbose = args.is_present("VERBOSE");
+            (subcommand, migration_dir, steps, verbose)
+        }
+        // Catch command `migrate`, this will be treated as `migrate up`
+        _ => {
+            let migration_dir = matches.value_of("MIGRATION_DIR").unwrap();
+            let verbose = matches.is_present("VERBOSE");
+            ("up", migration_dir, None, verbose)
+        }
+    };
+    // Construct the `--manifest-path`
+    let manifest_path = if migration_dir.ends_with('/') {
+        format!("{}Cargo.toml", migration_dir)
+    } else {
+        format!("{}/Cargo.toml", migration_dir)
+    };
+    // Construct the arguments that will be supplied to `cargo` command
+    let mut args = vec![
+        "run",
+        "--manifest-path",
+        manifest_path.as_str(),
+        "--",
+        subcommand,
+    ];
+    if let Some(steps) = steps {
+        args.extend(["-n", steps]);
+    }
+    if verbose {
+        args.push("-v");
+    }
+    // Run migrator CLI on user's behalf
+    println!("Running `cargo {}`", args.join(" "));
+    Command::new("cargo").args(args).spawn()?.wait()?;
+    Ok(())
+}
+
 fn handle_error<E>(error: E)
 where
     E: Display,
@@ -176,13 +273,14 @@ where
 
 #[cfg(test)]
 mod tests {
-    use clap::AppSettings;
-    use url::ParseError;
-
     use super::*;
+    use clap::AppSettings;
 
-    #[async_std::test]
-    async fn test_generate_entity_no_protocol() {
+    #[test]
+    #[should_panic(
+        expected = "called `Result::unwrap()` on an `Err` value: RelativeUrlWithoutBase"
+    )]
+    fn test_generate_entity_no_protocol() {
         let matches = cli::build_cli()
             .setting(AppSettings::NoBinaryName)
             .get_matches_from(vec![
@@ -192,22 +290,13 @@ mod tests {
                 "://root:root@localhost:3306/database",
             ]);
 
-        let result = std::panic::catch_unwind(|| {
-            smol::block_on(run_generate_command(matches.subcommand().1.unwrap()))
-        });
-
-        // Make sure result is a ParseError
-        match result {
-            Ok(Err(e)) => match e.downcast::<ParseError>() {
-                Ok(_) => (),
-                Err(e) => panic!("Expected ParseError but got: {:?}", e),
-            },
-            _ => panic!("Should have panicked"),
-        }
+        smol::block_on(run_generate_command(matches.subcommand().1.unwrap())).unwrap();
     }
 
     #[test]
-    #[should_panic]
+    #[should_panic(
+        expected = "There is no database name as part of the url path: postgresql://root:root@localhost:3306"
+    )]
     fn test_generate_entity_no_database_section() {
         let matches = cli::build_cli()
             .setting(AppSettings::NoBinaryName)
@@ -218,12 +307,13 @@ mod tests {
                 "postgresql://root:root@localhost:3306",
             ]);
 
-        smol::block_on(run_generate_command(matches.subcommand().1.unwrap()))
-            .unwrap_or_else(handle_error);
+        smol::block_on(run_generate_command(matches.subcommand().1.unwrap())).unwrap();
     }
 
     #[test]
-    #[should_panic]
+    #[should_panic(
+        expected = "There is no database name as part of the url path: mysql://root:root@localhost:3306/"
+    )]
     fn test_generate_entity_no_database_path() {
         let matches = cli::build_cli()
             .setting(AppSettings::NoBinaryName)
@@ -234,12 +324,11 @@ mod tests {
                 "mysql://root:root@localhost:3306/",
             ]);
 
-        smol::block_on(run_generate_command(matches.subcommand().1.unwrap()))
-            .unwrap_or_else(handle_error);
+        smol::block_on(run_generate_command(matches.subcommand().1.unwrap())).unwrap();
     }
 
     #[test]
-    #[should_panic]
+    #[should_panic(expected = "No username was found in the database url")]
     fn test_generate_entity_no_username() {
         let matches = cli::build_cli()
             .setting(AppSettings::NoBinaryName)
@@ -250,12 +339,11 @@ mod tests {
                 "mysql://:root@localhost:3306/database",
             ]);
 
-        smol::block_on(run_generate_command(matches.subcommand().1.unwrap()))
-            .unwrap_or_else(handle_error);
+        smol::block_on(run_generate_command(matches.subcommand().1.unwrap())).unwrap();
     }
 
     #[test]
-    #[should_panic]
+    #[should_panic(expected = "called `Result::unwrap()` on an `Err` value: PoolTimedOut")]
     fn test_generate_entity_no_password() {
         let matches = cli::build_cli()
             .setting(AppSettings::NoBinaryName)
@@ -266,12 +354,12 @@ mod tests {
                 "mysql://root:@localhost:3306/database",
             ]);
 
-        smol::block_on(run_generate_command(matches.subcommand().1.unwrap()))
-            .unwrap_or_else(handle_error);
+        smol::block_on(run_generate_command(matches.subcommand().1.unwrap())).unwrap();
     }
 
-    #[async_std::test]
-    async fn test_generate_entity_no_host() {
+    #[test]
+    #[should_panic(expected = "called `Result::unwrap()` on an `Err` value: EmptyHost")]
+    fn test_generate_entity_no_host() {
         let matches = cli::build_cli()
             .setting(AppSettings::NoBinaryName)
             .get_matches_from(vec![
@@ -281,17 +369,6 @@ mod tests {
                 "postgres://root:root@/database",
             ]);
 
-        let result = std::panic::catch_unwind(|| {
-            smol::block_on(run_generate_command(matches.subcommand().1.unwrap()))
-        });
-
-        // Make sure result is a ParseError
-        match result {
-            Ok(Err(e)) => match e.downcast::<ParseError>() {
-                Ok(_) => (),
-                Err(e) => panic!("Expected ParseError but got: {:?}", e),
-            },
-            _ => panic!("Should have panicked"),
-        }
+        smol::block_on(run_generate_command(matches.subcommand().1.unwrap())).unwrap();
     }
 }
